#!/usr/bin/python3

# vtable helpers
# this should probably be cleaned up and moved somewhere else

import collections
import itertools
import operator
import typing

import itanium_demangler as demangler
from cle.backends.symbol import Symbol

from . import demangler_helpers as dh
from .angr import VtableDisambiguator

if typing.TYPE_CHECKING:
    from .validate import LinuxBinary

VTable = list[Symbol]

# hotfix for demangler
demangler._is_ctor_or_dtor = demangler.is_ctor_or_dtor


def reorder_vfns_windows_estimate(symbols: list[Symbol], start_pos) -> list[Symbol]:
    # reorders a given subclass-level slice of linux symbols to reflect windows ordering
    name_buckets = collections.defaultdict(list)
    for n, symbol in enumerate(symbols):
        # collect overrides into buckets based on function name
        dmsym = demangler.parse(symbol.name)
        if dmsym:
            name_buckets[dh.extract_method_fname(dmsym)].append(symbol)
        else:
            # HACK: preserves positions of references to __cxa_pure_virtual
            name_buckets[(dmsym, n)].append(symbol)

    output_symbols: list[Symbol] = []
    for fname, syms in name_buckets.items():
        # on windows, overloads are made consecutive and in reverse of declared order
        # iteration order is guaranteed as of py3.7+ to be the insertion order,
        # so this should output symbols otherwise in their original order
        output_symbols.extend(reversed(syms))

    return output_symbols


def get_windows_vtables_from(
    bin: "LinuxBinary", vtda: VtableDisambiguator, vt: Symbol
) -> VTable:
    vt_typeinfo = bin.angr.loader.memory.unpack_word(vt.rebased_addr + 0x4)

    vt_parent_spans = [1]
    for typeinfo_ptr, name in reversed(list(vtda.dump_class_parents(vt_typeinfo))):
        vt_parent = bin.angr.loader.find_symbol(f"_ZTV{name}")
        if not vt_parent:
            continue
        vt_first, *_ = get_vtables_from_address(bin, vtda, vt_parent)
        vt_parent_spans.append(len(vt_first))

    vt_first, *vt_others = get_vtables_from_address(bin, vtda, vt)
    thunk_fns = set()
    for vt_other in vt_others:
        for sym in vt_other:
            dmsym = demangler.parse(sym.name)
            if dmsym.kind == "nonvirt_thunk":
                thunk_fns.add(dh.extract_method_signature(dmsym))

    vt_out = []
    for vt_low, vt_high in itertools.pairwise(vt_parent_spans):
        # we can only reorder overloads within the class they were initially specified
        # e.g. CTFPlayer's ChangeTeam cannot be merged with CBaseEntity's
        class_vfns = []

        for sym in vt_first[vt_low:vt_high]:
            # filter MI thunks
            dmsym = demangler.parse(sym.name)
            if dmsym:
                if (
                    not demangler.is_ctor_or_dtor(dmsym)
                    and dh.extract_method_signature(dmsym) in thunk_fns
                ):
                    continue
            else:
                # __cxa_pure_virtual returns None here; we still add it to the vtable slice
                pass
            class_vfns.append(sym)
        vt_out.extend(reorder_vfns_windows_estimate(class_vfns, vt_low))

    return vt_out


def get_vtables_from_address(
    bin: "LinuxBinary", vtda: VtableDisambiguator, vt: Symbol
) -> list[VTable]:
    # returns a list of vtables for each vtable present on the class
    VTableFunction = collections.namedtuple("VTableFunction", "tblidx sym")

    table_index = 0
    function_list: list[VTableFunction] = []
    addr_range = iter(range(vt.rebased_addr, vt.rebased_addr + vt.size, 4))
    for n, addr in enumerate(addr_range):
        # get symbols that map to that address
        deref = bin.angr.loader.fast_memory_load_pointer(addr)
        # assert deref
        fnsyms = set(vtda.syms_by_addr.get(deref) or set())
        if not fnsyms:
            # vtable boundary
            table_index += 1
            next(addr_range)  # consume typeinfo
            continue

        if len(fnsyms) == 1 or not function_list:
            function_list.append(VTableFunction(table_index, fnsyms.pop()))
            continue
        elif len(fnsyms) > 1:
            # function in vtable is referenced by multiple names; perform disambiguation
            matched_overload = None
            for related in vtda.get_possible_vtable_set_candidates(vt, n):
                matched_overload = vtda.resolve_ambiguous_vfn(n, fnsyms, related)
                if matched_overload:
                    break

            # it's possible that the other function(s) is/are resolveable.
            # without doing multiple passes and saving the disambiguity somewhere it'll be difficult to match

            if matched_overload:
                function_list.append(VTableFunction(table_index, matched_overload))
                continue
        # TODO: we need to resolve a symbol here; otherwise we can't correctly handle this
        vt_name = demangler.parse(vt.name)
        candidate_names = set(sym.name for sym in fnsyms)
        raise Exception(f"Ambiguity in {vt_name} position {n}; candidates {candidate_names}")

    return [
        list(vfn.sym for vfn in vtbl)
        for tblidx, vtbl in itertools.groupby(function_list, key=operator.attrgetter("tblidx"))
    ]
